#ifndef FC__PATH_FINDING_COMMON_H
#define FC__PATH_FINDING_COMMON_H

#include "map.h"
#include "unit.h"
#include "unittype.h"
#include "terrain.h"

/*
 * Functions in this file help to find a path from A to B.
 *
 * Terms used:
 *  - step: one movement step which brings us from one tile to an adjacent one
 *  - turn: obvious
 *  - path: a list of steps which leads from the start to the end
 *  - basic movement cost (BMC) (of a step): the basic movement costs
 *  of the step as returned by map_move_cost. Note that the BMC of a
 *  step is 0 if the target is unknown.
 *  - extra cost of a tile (ECOT): a user supplied value which is >=0
 *  - total extra cost (TEC): the sum of all ECOT of all tiles which
 *  are used
 *  - cost of a path (COP): set the comment of get_cop for details.
 *  - best path: a path which has the minimal COP. If two paths have
 *  equal COP the behavior is undefined.
 */

/*
 * The path finding functions can be used in two ways:
 *
 * A) the caller knows the goal position and wants to know the path:
 * 
 * struct pf_parameter parameter;
 * pf_map_t pf_map;
 * struct pf_path path;
 * 
 * // fill parameter
 *
 * pf_map = pf_get_map(&parameter);
 *
 * pf_get_path(pf_map, goal_x, goal_y, &path);
 *
 * if(path.is_valid) {
 *   // success
 * } else {
 *   // no path could be found
 * }
 *
 * pf_destroy_map(pf_map);
 *
 * You may call pf_get_path multiple times with the same pf_map.
 *
 * B) the caller doesn't know the goal yet and wants to iterate over
 * all paths in order of increasing costs (COP):
 *
 * struct pf_parameter parameter;
 * pf_map_t pf_map;
 * struct pf_path path;
 * 
 * // fill parameter
 *
 * pf_map = pf_get_map(&parameter);
 *
 * while (pf_next(pf_map)) {
 *   // get information about the current path
 *
 *   // do action (ignoring, saving, aborting the loop, ...) for this path
 * }
 *
 * pf_destroy_map(pf_map);
 *
 * Depending on the kind of information required the first part of the
 * while-body may look like:
 *
 *  1) the map position of the end of the path:
 *     int x = pf_next_get_x(pf_map);
 *     int y = pf_next_get_y(pf_map);
 *
 *  2) information (for example the total BMC or the number of turns)
 *  at the end of the path:
 *     struct pf_position pos;
 *     pf_next_get_position(pf_map, &pos);
 *
 *  3) information about the whole path:
 *     struct pf_path path;
 *     pf_next_get_path(pf_map, &path);
 */

/*
 * About filling the struct pf_parameter:
 *
 * This can either be done by hand or using the pft_* functions from
 * path_finding_tools or a mix of these.
 *
 * Note that it is useful to limit the expansion of unknown tiles with
 * the get_TB callback because the BMC of unknown tiles is 0.
 */

/*
 * Maximal number of steps in a path.
 */
#define MAX_PATH_LENGTH	200

#define PF_TURN_FACTOR	65536

enum tile_behavior {
  TB_IGNORE,
  TB_DONT_LEAVE,
  TB_NORMAL
};

/*
 * Specify which turn numbers and moves_left are used. TM_NONE will
 * set turn and moves_left to -1;
 */
enum turn_mode {
  TM_NONE, TM_AVERAGE, TM_LUCK, TM_SAFE
};

struct pf_path {
  /*
   * Is this path valid? Is set to FALSE if there was no path found.
   */
  bool is_valid;

  /*
   * Number of positions used in the array "positions".
   */
  int positions_used;

  struct pf_position {
    /*
     * moves_left and turn describe the arrival at the position (x, y).
     */
    int x, y;
    int turn, moves_left;
    int BMC_of_next_step;
    int cost;

    /* 
     * Total BMC required to reach this position. Undefined if a turn
     * mode other than TM_NONE is used.
     */
    int total_BMC;

    /* 
     * TEC required to reach this position.
     */
    int TEC;

    /* 
     * COP of the path which reaches this position.
     */
    int COP;

    enum direction8 direction_for_next_position;
  } positions[MAX_PATH_LENGTH + 1];
};

struct pf_parameter {
  /*
   * Definition of the state of a virtual unit unit at the start.
   */
  int start_x, start_y;
  int moves_left_initially;
  struct player *owner;

  enum unit_move_type move_type;
  /* only F_IGZOC and F_IGTER are used currently */
  bv_flags flags;
  int move_rate;

  /* 
   * Only affects the BMC. If set the FALSE the BMC is calculated to
   * simulate a path from (start_x, start_y) to the location (which is
   * returned by pf_get_next_position or passed to pf_get_path). If
   * set to TRUE the BMC will be calculated to simulate a path from
   * the location to the starting map position (start_x,start_y).
   */
  bool move_backward;

  bool ignore_enemy;
  bool omniscience;

  /*
   * Determines how turns and moves_left are calculated.
   */
  enum turn_mode turn_mode;

  /*
   * Callback to query the known state of a tile for the given
   * player. Should take into account the handicaps. If NULL
   * map_get_known2 is used.
   */
  enum known_type (*get_known) (int, int, struct player *);

  /*
   * GOTO_MOVE_STRAIGHTEST will set the BMC to constant 1. Even for
   * unknown tiles.
   */
  enum goto_move_restriction restriction;

  /*
   * Passed as last parameter to get_TB.
   */
  void *get_TB_data;

  /*
   * Callback which determines the behavior of a tile. If NULL
   * TB_NORMAL is assumed. Parameters are x, y, known and
   * get_TB_data. It can be assumed that the implementation of
   * path_finding.h will cache this value.
   */
  enum tile_behavior (*get_TB) (int, int, enum known_type, void *);

  /*
   * Passed as last parameter to get_ECOT.
   */
  void *get_ECOT_data;

  /*
   * Callback which can be used to provide extra costs depending on
   * the tile. Can be NULL. Parameters are x, y, known and
   * get_ECOT_data. It can be assumed that the implementation of
   * path_finding.h will cache this value.
   */
  int (*get_ECOT) (int, int, enum known_type, void *);

  /*
   * Passed as last parameter to is_position_safe.
   */
  void *is_position_dangerous_data;

  /*
   * If this callback is non-NULL and returns TRUE this position is
   * dangerous. The unit will never end a turn at a dangerous
   * position. Can be NULL. Parameter are x, y and
   * is_position_dangerous_data.
   */
  bool(*is_position_dangerous) (int, int, void *);
  
  bool(*is_pos_dangerous) (int, int, void *);
  void *is_pos_dangerous_data;
  bool zoc_used;
};

#endif
