#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
/* TODO: For testing period only.  Remove! */
#include <time.h>


#include "combat.h"
#include "game.h"
#include "log.h"
#include "map.h"
#include "mem.h"
#include "rand.h"

#include "path_finding_gb.h"

/* For speed set to powers of 2
 * Or not: large chunks of memory are verrrry slow */
#define MAX_XSIZE map.xsize
#define MAX_YSIZE map.ysize

/* Large enough */
#define PF_MAXCOST 2000


/*======================================================================*/
/*===================== Hidden structures ==============================*/
/*======================================================================*/
/*
 * xy_t is the offset into the array of nodes
 */
typedef short xy_t;
typedef signed char stiny_t;
typedef unsigned char utiny_t;

/*
 * Some comments on implementation: 
 * 1. cost is sum of BMCs altered to fit to the turn_mode
 * see update_costs
 * 2. dir_to_here is for backtracking along the tree of shortest paths
 * 3. node_known_type, behavior, is_my_zoc, is_allied_tile, extra
 * are all cached values.  It is possible to shove them into a separate 
 * array which is allocated only if a corresponding option in the
 * parameter is set.
 */
struct pf_node {
  int cost;                       /* total cost up to now */
  int extra_cost;                 /* total extra cost up to now */
  utiny_t dir_to_here;
  utiny_t node_known_type;
  utiny_t behavior;
  utiny_t is_my_zoc;
  utiny_t is_allied_tile;
  int extra_tile;                 /* extra cost of the tile */
};
typedef struct pf_node *pf_node_t;

/* 
 * All danger-related fields go into here.  For speed we separate it
 * from mainstream node.
 */
struct danger_node {
  bool is_dangerous;
  bool waited;                 /* TRUE if waited to get here */
  struct segment_node {
    enum direction8 dir;
    int cost;
    int extra_cost;
  } *danger_segment;           /* Danger segment 
                                * need to remeber costs and stuff */
};

enum pf_node_status {
  NS_UNINIT = 0,                /* memory is calloced, hence zero 
                                 * means uninitialised */
  NS_NEW,                       /* the optimal route isn't found yet */
  NS_WAITING,                   /* the optimal route is found,
                                 * considering waiting */
  NS_PROCESSED                  /* the optimal route is found */
};

/* 
 * The queue will probably be separated into another module
 */
#define PQDATUM xy_t
#define PQDATUM_NULL -1
#define INIT_ITEMS 100

struct pqueue {
  int size;        /* number of occupied cells */
  int avail;       /* total number of cells */
  int step;        /* additional memory allocation step */
  PQDATUM *cells;  /* array containing data */
  int *priorities; /* backup priorities (in case data is changed 
                    * while in the queue) */
};


/* 
 * Generic move cost function:
 * (x,y) is the target tile, dir is the direction to the target tile.
 */
typedef int (COSTFN)(pf_map_t the_map, int x, int y, int dir);

/*
 * The map structure itself.
 */
struct pf_map {
  short x, y;                      /* The current position */
  xy_t xy;                         /* Current offset into lattice */
  COSTFN *get_cost;                /* The move cost function */
  struct pf_parameter *params;
  struct pqueue *queue;
  struct pf_node *lattice;
  utiny_t *status;                   /* Array of node stati 
                                    * (enum pf_node_status really) */
  struct pqueue *danger_queue;     /* Dangerous positions go there */
  struct danger_node *d_lattice;   /* Lattice with danger stuff */
};

bool pf_danger_iterate_map(pf_map_t the_map);
bool pf_danger_get_current_path(const pf_map_t the_map, 
                                struct pf_path *path);


/* ------------------------ The queue implementation ----------------- */

/**********************************************************************
 * pqinit: initialize the queue.
 *
 * n is the numer of queue items for which memory should be 
 * preallocated, that is, the initial size of the item array the queue 
 * uses. If you insert more than n items to the queue, another n items 
 * will be allocated automatically.
 **********************************************************************/
struct pqueue *pqcreate(int n)
{
  struct pqueue *q;

  q = fc_malloc(sizeof(struct pqueue));
  q->cells = fc_malloc(sizeof(PQDATUM) * n);
  q->priorities = fc_malloc(sizeof(int) * n);
  q->avail = q->step = n;
  q->size = 1;
  return q;
}


/********************************************************************
 * Destructor for queue structures
 *******************************************************************/
void pqdestroy(struct pqueue *q)
{
  free(q->cells);
  free(q->priorities);
  free(q);
}


/********************************************************************
 * pqinsert: insert an item into the queue.
 *
 * q is the pointer to a priority queue.
 * d is the datum to be inserted.
 ********************************************************************/
void pqinsert(struct pqueue *q, PQDATUM datum, int datum_priority)
{
  int i, newsize;

  assert(q);

  /* allocate more memory if necessary */
  if (q->size >= q->avail) {
    PQDATUM *tmp_d;
    int *tmp_p;
    newsize = q->size + q->step;
    freelog(LOG_NORMAL, "Resizing queue to %d", newsize);
    tmp_d = fc_realloc(q->cells, sizeof(PQDATUM) * newsize);
    tmp_p = fc_realloc(q->priorities, sizeof(int) * newsize);

    q->cells = tmp_d;
    q->priorities = tmp_p;
    q->avail = newsize;
  }

  /* insert item */
  i = q->size++;
  while (i > 1 && q->priorities[i / 2] < datum_priority) {
    q->cells[i] = q->cells[i / 2];
    q->priorities[i] = q->priorities[i / 2];
    i /= 2;
  }
  q->cells[i] = datum;
  q->priorities[i] = datum_priority;
}


/*******************************************************************
 * pqremove: remove the highest-ranking item from the queue.
 *
 * p is the pointer to a priority queue.
 *
 * Return value:
 *    non-NULL    The value of the item that has been removed.
 *    NULL        No item could be removed. Either the queue pointer
 *                provided was NULL, or the queue was empty.
 *******************************************************************/
PQDATUM pqremove(struct pqueue * q)
{
  PQDATUM tmp;
  int tmp_priority;
  PQDATUM top;
  int i = 1, j;

  assert(q->size <= q->avail);
  if (!q || q->size == 1)
    return PQDATUM_NULL;
  top = q->cells[1];
  q->size--;
  tmp = q->cells[q->size];
  tmp_priority = q->priorities[q->size];
  while (i <= q->size / 2) {
    j = 2 * i;
    if (j < q->size && q->priorities[j] < q->priorities[j + 1]) {
      j++;
    }
    if (q->priorities[j] <= tmp_priority) {
      break;
    }
    q->cells[i] = q->cells[j];
    q->priorities[i] = q->priorities[j];
    i = j;
  }
  q->cells[i] = tmp;
  q->priorities[i] = tmp_priority;
  return top;
}


/* ------------ The conversion methods (x,y) <--> xy ------------ */

/*******************************************************************
 * (x,y) --> xy
 *******************************************************************/
#define get_xy_from_x_and_y(x, y) (y * MAX_XSIZE + x)

/*******************************************************************
 * xy --> x
 *******************************************************************/
#define get_x_from_xy(xy) (xy % MAX_XSIZE)

/*******************************************************************
 * xy --> y
 *******************************************************************/
#define get_y_from_xy(xy) (xy / MAX_XSIZE)



/*-------------------- Possible cost functions -------------------*/

/*************************************************************
 * SINGLE_MOVE cost function 
 *************************************************************/
int single_move(pf_map_t the_map, int x, int y, int dir)
{
  return SINGLE_MOVE;
}

/*************************************************************
 * SINGLE_MOVE cost function for SEA_MOVING 
 * Note: this is a reversable function
 *************************************************************/
int single_seamove(pf_map_t the_map, int x, int y, int dir)
{
  /* -3 means ships can move between */
  if ( map_get_tile(the_map->x, the_map->y)->move_cost[dir] 
       == MOVE_COST_FOR_VALID_SEA_STEP) {
    return SINGLE_MOVE;
  } else {
    return PF_MAXCOST;
  }
}

/************************************************************ 
 * LAND_MOVE cost function for a unit 
 ************************************************************/
int normal_move_unit(pf_map_t the_map, int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(the_map->x, the_map->y);
  int t1 = map_get_terrain(x1, y1);
  int move_cost;

  if (t1 == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0) {
      move_cost = SINGLE_MOVE;
    } else {
      move_cost = PF_MAXCOST;
    }
  } else if (ptile->terrain == T_OCEAN) {
    move_cost = get_tile_type(t1)->movement_cost * SINGLE_MOVE;
  } else {
    move_cost = ptile->move_cost[dir];
  }

  return move_cost;
}

/************************************************************ 
 * Reveresed  !!
 * LAND_MOVE cost function for a unit 
 ************************************************************/
int reverse_move_unit(pf_map_t the_map, int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(x1, y1);
  int ter0 = map_get_terrain(the_map->x, the_map->y);
  int ter1 = ptile->terrain;
  int move_cost = PF_MAXCOST;

  if (ter1 == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0) {
      /* Landing */
      move_cost = get_tile_type(ter0)->movement_cost * SINGLE_MOVE;
    } else {
      /* Nothing to land from */
      move_cost = PF_MAXCOST;
    }
  } else if (ter0 == T_OCEAN) {
    /* Boarding */
    move_cost = SINGLE_MOVE;
  } else {
    move_cost = ptile->move_cost[DIR_REVERSE(dir)];
  }

  return move_cost;
}

/************************************************************ 
 * IGTER_MOVE cost function for a unit 
 ************************************************************/
static int igter_move_unit(pf_map_t the_map, int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(the_map->x, the_map->y);
  int move_cost;

  if (map_get_terrain(x1, y1) == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0) {
      move_cost = MOVE_COST_ROAD;
    } else {
      move_cost = PF_MAXCOST;
    }
  } else if (ptile->terrain == T_OCEAN) {
    move_cost = MOVE_COST_ROAD;
  } else {
    move_cost = (ptile->move_cost[dir] ? MOVE_COST_ROAD : 0);
  }
  return move_cost;
}

/************************************************************ 
 * Reversed !! 
 * IGTER_MOVE cost function for a unit 
 ************************************************************/
static int reverse_igter_move_unit(pf_map_t the_map, 
                                   int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(x1, y1);
  int move_cost;

  if (map_get_terrain(x1, y1) == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0) {
      /* Landing */
      move_cost = MOVE_COST_ROAD;
    } else {
      move_cost = PF_MAXCOST;
    }
  } else if (map_get_terrain(the_map->x, the_map->y) == T_OCEAN) {
    /* Boarding */
    move_cost = MOVE_COST_ROAD;
  } else {
    move_cost = (ptile->move_cost[DIR_REVERSE(dir)] ? MOVE_COST_ROAD : 0);
  }
  return move_cost;
}

#if 0 
/*************************************************************
 * LAND_MOVE cost function for a city ?? 
 * NOT TESTED.  USED?
 *************************************************************/
static int normal_move_city(pf_map_t the_map, int x1, int y1, int dir)
{
  /* have no idea what code below is for */

  if (map_get_terrain(x1, y1) == T_OCEAN)
    return PF_MAXCOST;
  else {
    struct tile *ptile = map_get_tile(the_map->x, the_map->y);
    int tmp = map_get_tile(x1, y1)->move_cost[DIR_REVERSE(dir)];
    return (ptile->move_cost[dir] + tmp +
	    (ptile->move_cost[dir] > tmp ? 1 : 0))/2;
  }
}
#endif


/* --------------------- extra cost functions --------------------- */

/*********************************************************************
 * An example of extra cost fn: will avoid plains if possible 
 *********************************************************************/
int avoid_plains(int x, int y, enum known_type known, void *data)
{
  enum tile_terrain_type ter = map_get_terrain(x, y);

  switch (ter){
  case T_PLAINS:
    return 3;
  default:
    return 0;
  }
}

/* -------------------- tile status functions --------------------- */

/*********************************************************************
 * Will not go into mountains and will not pass through hills
 *********************************************************************/
enum tile_behavior avoid_hills_TB(int x, int y, 
                                  enum known_type known, void *data) 
{
  switch (map_get_terrain(x, y)){
  case T_HILLS:
    return TB_DONT_LEAVE;
  case T_MOUNTAINS:
    return TB_IGNORE;
  default:
    return TB_NORMAL;
  }
}

/*********************************************************************
 * Strict trireme function: will never go into the unknown and 
 * away from shore.
 *********************************************************************/
enum tile_behavior strict_trireme(int x, int y, 
                                  enum known_type known, void *data)
{
  if (known != TILE_UNKNOWN && is_coastline(x,y)) {
    return TB_NORMAL;
  }
  return TB_IGNORE;
}

/*********************************************************************
 * Just a test thing, for danger walking
 *********************************************************************/
int avoid_20(int x, int y, enum known_type known, void *data)
{
  return (x >= 20 ? 6000 : 0); 
}


/* -------------------- position dengerous functions -------------- */

/********************************************************************
 * Will consider grassland to be dangerous (ooh, grass, spooky!)
 ********************************************************************/
bool grass_is_spooky(int x, int y, void *data)
{
  return (map_get_terrain(x, y) == T_GRASSLAND);
}   


/* -------------------- manipulating the cost --------------------- */

/********************************************************************
 * Number of turns requite to reach node
 ********************************************************************/
#define get_turn(map, cost) \
  ((cost) / (map)->params->move_rate)

/********************************************************************
 * Moves left after node is reached
 ********************************************************************/
#define get_moves_left(map, cost) \
  ((map)->params->move_rate \
   - ((cost) % (map)->params->move_rate))

/********************************************************************
 * Adjust BMC to reflect the turn mode and the move_rate.
 * The result is AMC.
 ********************************************************************/
#define adjust_cost(map, cost) \
  if (cost < PF_MAXCOST) { \
    switch (map->params->turn_mode) { \
    case TM_NONE: \
      break; \
    case TM_AVERAGE: \
      cost = MIN(cost, map->params->move_rate); \
      break; \
    case TM_SAFE: \
      { \
        int moves_left = get_moves_left(map, map->lattice[map->xy].cost); \
        \
        cost = (cost > moves_left ? moves_left + cost : cost); \
        break; \
      } \
    case TM_LUCK: \
      { \
        int moves_left = get_moves_left(map, map->lattice[map->xy].cost); \
        \
        cost = (cost > moves_left ? moves_left : cost); \
        break; \
      } \
    } \
  }


#if 0
/* ------------------------ get_known function(s) ----------------- */
/********************************************************************
 * Server-side get_known function 
 ********************************************************************/
enum known_type server_get_known(int x, int y, struct player *pplayer)
{
  if (!map_get_known(x, y, pplayer)) {
    return TILE_UNKNOWN;
  } else if (map_get_known_and_seen(x, y, pplayer)) { 
    return TILE_KNOWN;
  } else {
    return TILE_KNOWN_FOGGED;
  }
}
#endif


/* ---------------------- Path-finding proper --------------------*/

/******************************************************************
 * Calculates cached values of the target node: 
 * node_known_type and zoc
 ******************************************************************/
static void pf_init_node(pf_map_t the_map, pf_node_t the_node, 
                         int x, int y)
{
  struct pf_parameter *params = the_map->params;

  /* We will change the status of the tile once we have put
   * sensible values into the_node->cost */

  /* Establish the "known" status of the_node */
  if (params->get_known) {
    the_node->node_known_type 
      = (*(params->get_known))(x, y, params->owner);
  } else {
    the_node->node_known_type = TILE_KNOWN;
  }

  /* Establish the tile behavior */
  if (params->get_TB) {
    the_node->behavior
      = (*(params->get_TB))(x, y, the_node->node_known_type, 
                            params->get_TB_data);
  } else {
    /* The default */
    the_node->behavior = TB_NORMAL;
  }

  if (the_map->params->zoc_used) {
    struct tile *the_tile = map_get_tile(x, y);
    
    the_node->is_my_zoc = (the_tile->city
                           || the_tile->terrain == T_OCEAN
                           || is_my_zoc(params->owner, x, y));
    the_node->is_allied_tile =
      (is_allied_unit_tile(the_tile, params->owner) != NULL);
  }

  /* Evaluate the extra cost of the destination */
  if (params->get_ECOT) {
    the_node->extra_tile 
      = (*(params->get_ECOT))(x, y, the_node->node_known_type, 
                              params->get_ECOT_data);
  } else {
    the_node->extra_tile = 0;
  }

}


/* -------------------- generating the map ------------------------ */

/*****************************************************************
 * Primary method for iterative path-finding.
 * Plan: 1. Process previous position
 *       2. Get new nearest position and return it
 *****************************************************************/
bool pf_next(pf_map_t the_map)
{
  xy_t xy;
  pf_node_t node_xy = &the_map->lattice[the_map->xy];

#define COP(cost, extra) ( PF_TURN_FACTOR * (cost) \
                           + (extra) * (the_map->params)->move_rate )

  if (the_map->params->is_pos_dangerous) {
    /* It's a lot different if is_pos_dangerous is defined */
    return pf_danger_iterate_map(the_map);
  }

  the_map->status[the_map->xy] = NS_PROCESSED;

  /* There is no exit from NOTHROUGH tiles! */
  if (node_xy->behavior != TB_DONT_LEAVE) {

    /* Processing Stage */
    /* The previous position is contained in {x,y} fields of the_map */
    
    adjc_dir_iterate(the_map->x, the_map->y, x1, y1, dir) {
      xy_t xy1 = get_xy_from_x_and_y(x1, y1);
      pf_node_t node_xy1 = &the_map->lattice[xy1];
      utiny_t *status = &the_map->status[xy1];
      int cost;
      int extra = 0;
      
      if (*status == NS_PROCESSED) {
        /* This gives 15% speedup */
        continue;
      }

      if (*status == NS_UNINIT) {
        pf_init_node(the_map, node_xy1, x1, y1);
      }
      
      /* Can we enter this tile at all? */
      if(node_xy1->behavior == TB_IGNORE) {
        continue;
      }

      /* Is the move ZOC-ok? */
      if (the_map->params->zoc_used 
          && !(node_xy->is_my_zoc || node_xy1->is_my_zoc
               || node_xy1->is_allied_tile)) {
        continue;
      }
          
      /* Evaluate the cost of the move */
      cost = (*(the_map->get_cost))(the_map, x1, y1, dir);
      adjust_cost(the_map, cost);
      
      /* assert( cost >= 0 );  No special cases yet */
      
      /* Total cost at xy1 */
      cost += node_xy->cost;
      /* Is the cost reasonable? */
      if (cost >= PF_MAXCOST) {
        continue;
      } 
      
      /* Evaluate the extra cost of the destination,
       * if it's relevant */
      if (the_map->params->get_ECOT) {
        extra = node_xy1->extra_tile;
        extra += node_xy->extra_cost;
      }
      
      /* Update costs and add to queue, if we found a better route to xy1. */
      {
        int cost_of_path = COP(cost, extra);
        
        if (*status == NS_UNINIT
            || cost_of_path < COP(node_xy1->cost, node_xy1->extra_cost)) {
          *status = NS_NEW;
          node_xy1->extra_cost = extra;
          node_xy1->cost = cost;
          node_xy1->dir_to_here = dir;
          pqinsert(the_map->queue, xy1, -cost_of_path);
        }
      }
      
    } adjc_dir_iterate_end;
  }

  /* Get the next nearest node */
  do {
    xy = pqremove(the_map->queue);
    /* Discard if this node has already been processed */
  } while (xy != PQDATUM_NULL && the_map->status[xy] != NS_NEW);

  if (xy == PQDATUM_NULL) {
    return FALSE;
  }
  the_map->xy = xy;
  the_map->x = get_x_from_xy(xy);
  the_map->y = get_y_from_xy(xy);

  return TRUE;

#undef COP
}

/******************************************************************
 * Allocates the memory for the map.  No initialization.
 ******************************************************************/
pf_map_t pf_create_map()
{
  pf_map_t the_map = fc_calloc(1, sizeof(struct pf_map));

  the_map->lattice = 
    fc_malloc(MAX_XSIZE * MAX_YSIZE * sizeof(struct pf_node));
  the_map->queue = pqcreate(INIT_ITEMS);
  the_map->status = fc_calloc(MAX_XSIZE * MAX_YSIZE, 
                              sizeof(*(the_map->status)));

  return the_map;
}

/*****************************************************************
 * Create and fill the default parameter structure for a unit
 *****************************************************************/
struct pf_parameter *pf_create_unit_params(struct unit *punit)
{
  struct pf_parameter *params = fc_calloc(1, sizeof(struct pf_parameter));

  params->move_rate = unit_type(punit)->move_rate;
  params->start_x = punit->x;
  params->start_y = punit->y;
  params->owner = unit_owner(punit);
  params->moves_left_initially = punit->moves_left;
  params->move_type = unit_type(punit)->move_type;
  params->flags = unit_type(punit)->flags;

  return params;
}

/*****************************************************************
 * Creates the map and does minimal initializations.  
 * For debugging period only.  
 * TODO: Move it into pf_tools!
 *****************************************************************/
pf_map_t pf_init_map(struct unit *punit)
{
  struct pf_parameter *params = pf_create_unit_params(punit);
  pf_map_t the_map;
  
  params->turn_mode = TM_SAFE;
  params->get_known = NULL;
  params->zoc_used = FALSE;
  
  the_map = pf_get_map(params);

  return the_map;
}

/***************************************************************
 * Sets up the map according to the parameters
 * Does not do any iterations
 ***************************************************************/
pf_map_t pf_get_map(const struct pf_parameter *const parameter)
{
  pf_map_t the_map = pf_create_map();

  the_map->params = fc_malloc(sizeof(struct pf_parameter));
  *the_map->params = *parameter;

  the_map->x = the_map->params->start_x;
  the_map->y = the_map->params->start_y;
  the_map->xy = get_xy_from_x_and_y(the_map->x, the_map->y);

  switch (the_map->params->move_type) {
  case LAND_MOVING:
    if (BV_ISSET(the_map->params->flags, F_IGTER)) {
      the_map->get_cost = (the_map->params->move_backward 
                           ? reverse_igter_move_unit
                           : igter_move_unit);
    } else {
      the_map->get_cost = (the_map->params->move_backward 
                           ? reverse_move_unit
                           : normal_move_unit);
    }
    break;
  case SEA_MOVING:
    /* Symmetric wrt reversal */
    the_map->get_cost = single_seamove;
    if (BV_ISSET(the_map->params->flags, F_TRIREME)) {    
      the_map->params->get_TB = strict_trireme;
    }
  default:
    assert(0);
  }

  pf_init_node(the_map, &the_map->lattice[the_map->xy], 
               the_map->x, the_map->y);

  if (the_map->params->is_pos_dangerous) {
    /* Initialize stuff for dangerous positions */
    the_map->danger_queue = pqcreate(INIT_ITEMS);
    the_map->d_lattice = 
      fc_calloc(MAX_XSIZE * MAX_YSIZE, sizeof(struct danger_node));
    the_map->d_lattice[the_map->xy].is_dangerous = FALSE;
  } /* Otherwise they stay NULL */

  the_map->lattice[the_map->xy].cost = the_map->params->move_rate 
    - the_map->params->moves_left_initially;
  the_map->lattice[the_map->xy].extra_cost = 0;

  return the_map;
}


/*********************************************************************
 * After usage the map must be destroyed.
 *********************************************************************/
void pf_destroy_map(pf_map_t the_map)
{
  free(the_map->lattice);
  free(the_map->params);
  pqdestroy(the_map->queue);

  /* Danger-related structs */
  if (the_map->danger_queue) {
    pqdestroy(the_map->danger_queue);
  }
  
  if (the_map->d_lattice) {
    int i;

    /* Need to clean up the dangling danger_sements */
    for(i = 0; i < MAX_XSIZE * MAX_YSIZE; i++) {
      if (the_map->d_lattice[i].danger_segment) {
        free(the_map->d_lattice[i].danger_segment);
      }
    }
    free(the_map->d_lattice);
  }

  free(the_map);
}


/*---------------------- Lifting info from the map ----------------*/

/*******************************************************************
 * x of the current node
 *******************************************************************/
int pf_next_get_x(const pf_map_t the_map)
{
  return the_map->x;
}

/*******************************************************************
 * y of the current node
 *******************************************************************/
int pf_next_get_y(const pf_map_t the_map)
{
  return the_map->y;
}

/*******************************************************************
 * cost of the shortest path to the current node
 *******************************************************************/
int pf_get_current_cost(const pf_map_t the_map)
{
  return the_map->lattice[the_map->xy].cost;
}

/*******************************************************************
 * extra cost of the shortest path to the current node
 *******************************************************************/
int pf_get_current_extra_cost(const pf_map_t the_map)
{
  return the_map->lattice[the_map->xy].extra_cost;
}

/*******************************************************************
 * Read off the path to the current node
 *******************************************************************/
bool pf_next_get_path(const pf_map_t the_map, struct pf_path *path)
{
  struct pf_position path_rev[MAX_PATH_LENGTH + 1];
  int i;
  int x = the_map->x, y = the_map->y;
  enum direction8 dir_next = -1;

  if (the_map->params->is_pos_dangerous) {
    /* If is_pos_dangerous is defined, it's very different */
    return pf_danger_get_current_path(the_map, path);
  }

  for (i = 0; i <= MAX_PATH_LENGTH; i++) {
    pf_node_t node_xy = &the_map->lattice[get_xy_from_x_and_y(x, y)];
    path_rev[i].x = x;
    path_rev[i].y = y;
    path_rev[i].cost = node_xy->cost;
    path_rev[i].TEC = node_xy->extra_cost;
    if (the_map->params->turn_mode == TM_SAFE ||
        the_map->params->turn_mode == TM_LUCK) {
      path_rev[i].turn = get_turn(the_map, node_xy->cost);
      path_rev[i].moves_left = get_moves_left(the_map, node_xy->cost);
    } else {
      path_rev[i].turn = -1;
      path_rev[i].moves_left = -1;
    }

    path_rev[i].direction_for_next_position = dir_next;
    dir_next = node_xy->dir_to_here;

    if (same_pos(x, y, 
                 the_map->params->start_x, the_map->params->start_y)) {
      /* Ah-ha, reached the starting point! */
      int j;

      path->positions_used = i + 1;
      path->is_valid = TRUE;
      
      /* Actually filling the path */
      for(j = 0; i>=0; i--, j++) {
        memcpy(&path->positions[j], &path_rev[i], 
               sizeof(struct pf_position));
      }
      
      return TRUE;
    }

    /* Need to step further back */
    {
      int dummy_x, dummy_y;
      
      /* Ugly MAPSTEP... */
      if (!MAPSTEP(dummy_x, dummy_y, x, y, DIR_REVERSE(dir_next))) {
        freelog(LOG_NORMAL, "pf_get_current_path: wrong directions recorded!");
        path->is_valid = FALSE;
        return FALSE;
      } else {
        x = dummy_x;
        y = dummy_y;
      }
    }

  }

  freelog(LOG_NORMAL, "pf_get_current_path: path too long");
  path->is_valid = FALSE;
  return FALSE;
}

/************************************************************************
 * Printing a path
 ************************************************************************/
void pf_print_path(const struct pf_path *path)
{
  int i;

  freelog(LOG_NORMAL, "PF: path (at %p) is %s", path,
	  path->is_valid ? "valid" : "invalid");
  if (!path->is_valid) {
    return;
  }
  freelog(LOG_NORMAL, "PF:  consists of %d positions:",
	  path->positions_used);
  
  for (i = 0; i < path->positions_used; i++) {
    freelog(LOG_NORMAL,
	    "PF:   %2d/%2d: (%2d,%2d) cost=%2d (%2d, %d) EC=%d",
	    i + 1, path->positions_used, 
            path->positions[i].x, path->positions[i].y,
	    path->positions[i].cost, path->positions[i].turn, 
            path->positions[i].moves_left, path->positions[i].TEC);
  }
}

#if 0
/*********************************************************************
 * Time/test the path-finding execution.
 *********************************************************************/
void time_path_finding(struct player *pplayer, struct unit *punit)
{
  /*  struct tms *start, *finish;  */
  clock_t start, finish;
  int i;

  start = clock();

#define RUNS 1

  for(i = 0; i < RUNS; i++) {
    struct unit *aunit = punit;
    //    unit_list_iterate(pplayer->units, aunit) {
      struct pf_parameter *params = pf_create_unit_params(aunit);
      pf_map_t test_map;
      
      params->turn_mode = TM_SAFE;
      params->get_known = server_get_known;
      params->zoc_used = FALSE;
      params->get_ECOT = avoid_20;
      params->is_pos_dangerous = grass_is_spooky;
      
      test_map = pf_get_map(params);
      freelog(LOG_DEBUG, "%s (%d,%d)", 
              unit_type(aunit)->name, aunit->x, aunit->y);
      while (pf_iterate_map(test_map)) {
        freelog(LOG_DEBUG, "Processing (%d,%d): cost=%d, extra=%d",
                pf_get_current_x(test_map), pf_get_current_y(test_map),
                pf_get_current_cost(test_map), 
                pf_get_current_extra_cost(test_map));
        if (pf_get_current_x(test_map) == 21
            && pf_get_current_y(test_map) == 40) {
          struct pf_path *pathto2038
            = fc_malloc(sizeof(struct pf_path));
          pf_get_current_path(test_map, pathto2038);
          pf_print_path(pathto2038);
        }
      }
      pf_destroy_map(test_map);
      free(params);
      //    } unit_list_iterate_end;
  }

#undef RUNS

  finish = clock();

  freelog(LOG_NORMAL, "Secs used: %f", 
          (float) (finish - start) / CLOCKS_PER_SEC );

}
#endif

/* ----------------- Danger-related stuff ----------------------- */

/******************************************************************
 * Calculates cached danger-related values of the target node
 ******************************************************************/
static void init_danger_node(pf_map_t the_map, 
                             struct danger_node *d_node,
                             pf_node_t the_node, 
                             int x, int y)
{
  struct pf_parameter *params = the_map->params;

  /* Is the tile dangerous (i.e. no ending turn there) */
  if (params->is_pos_dangerous) {
    d_node->is_dangerous
      = (*(params->is_pos_dangerous))(x, y, 
                                      params->is_pos_dangerous_data);
  } else {
    freelog(LOG_ERROR, "PF: init_danger_node called without
                        is_pos_dangerous callback");
  }

}

/*****************************************************************
 * Creating path segment going back to a safe tile
 * TODO: Do dynamic (re)allocation of paths array!
 *****************************************************************/
struct segment_node *create_danger_segment(pf_map_t the_map, 
                                           enum direction8 dir,
                                           bool *waited)
{
  struct segment_node *path;
  int i = 1;
  int x = the_map->x;
  int y = the_map->y;
  pf_node_t node = &the_map->lattice[the_map->xy];
  struct danger_node *d_node = &the_map->d_lattice[the_map->xy];

  path = fc_malloc(MAX_PATH_LENGTH * sizeof(struct segment_node));

  path[0].dir = dir;
  path[0].cost = -1;
  path[0].extra_cost = -1;
  do {
    /* Record the direction */
    path[i].dir = node->dir_to_here;
    path[i].cost = node->cost;
    path[i].extra_cost = node->extra_cost;
    *waited = d_node->waited;
    /* Step further down the tree */
    {
      int dummy_x, dummy_y;
      
      /* Ugly MAPSTEP... */
      if (!MAPSTEP(dummy_x, dummy_y, x, y, DIR_REVERSE(node->dir_to_here))) {
        freelog(LOG_NORMAL, "create_extended_dir: wrong directions recorded!");
        assert(0);
      } else {
        x = dummy_x;
        y = dummy_y;
      }
    }
    node = &the_map->lattice[get_xy_from_x_and_y(x, y)];
    d_node = &the_map->d_lattice[get_xy_from_x_and_y(x, y)];
    i++;
  } while (i < MAX_PATH_LENGTH - 1 && d_node->is_dangerous);

  if (!d_node->is_dangerous) {
    /* Reached a safe (blue) node */
    path[i].dir = -1;  /* -1 meaning end of path */
  } else {
    freelog(LOG_ERROR, "PF: Unsafe segment too long!");
    free(path);
    return NULL;
  }

  return path;
}

/*****************************************************************
 * Primary method for iterative path-finding in presence of danger
 * Notes: 
 * 1. Whenever the path-finding stubles upon a dangerous 
 * location, it goes into a sub-Dijkstra which processes _only_ 
 * dangerous locations, by means of a separate queue.  When this
 * sub-Dijkstra reaches a safe location, it records the segment of
 * the path going across the dangerous terrain.
 * 2. Waiting is realised by inserting the (safe) tile back into 
 * the queue with a lower priority P.  This tile might pop back 
 * sooner than P, because there might be several copies of it in 
 * the queue already.  But that does not seem to present any 
 * problems.
 * 3. This algorithm cannot guarantee the best safe segments across 
 * dangerous region.  However it will find a safe segment if there 
 * is one.  To gurantee the best (in terms of COP) safe segments 
 * across danger, supply get_ECOT, which returns small extra on 
 * dangerous tiles.
 *****************************************************************/
bool pf_danger_iterate_map(pf_map_t the_map)
{
  xy_t xy;
  pf_node_t node_xy = &the_map->lattice[the_map->xy];
  struct danger_node *d_node_xy = &the_map->d_lattice[the_map->xy];
 
#define COP(cost, extra) ( PF_TURN_FACTOR * (cost) \
                           + (extra) * (the_map->params)->move_rate )
 
  /* There is no exit from NOTHROUGH tiles! */
  if (node_xy->behavior != TB_DONT_LEAVE) {
    /* Cost at xy but taking into account waiting */
    int loc_cost 
      = (the_map->status[the_map->xy] != NS_WAITING ? node_xy->cost
         : node_xy->cost + get_moves_left(the_map, node_xy->cost));
    
    /* The previous position is contained in {x,y} fields of the_map */    
    adjc_dir_iterate(the_map->x, the_map->y, x1, y1, dir) {
      xy_t xy1 = get_xy_from_x_and_y(x1, y1);
      pf_node_t node_xy1 = &the_map->lattice[xy1];
      struct danger_node *d_node_xy1 = &the_map->d_lattice[xy1];
      int cost;
      int extra = 0;
      
      /* Dangerous tiles can be uptaded even after being processed */
      if (the_map->status[xy1] == NS_PROCESSED 
          && !d_node_xy1->is_dangerous) {
        continue;
      }
      
      /* Initialise target tile if necessary */
      if (the_map->status[xy1] == NS_UNINIT) {
        pf_init_node(the_map, node_xy1, x1, y1);
        init_danger_node(the_map, d_node_xy1, node_xy1, x1, y1);
      }
      
      /* Can we enter this tile at all? */
      if (node_xy1->behavior == TB_IGNORE) {
        continue;
      }
      
      /* Is the move ZOC-ok? */
      if (the_map->params->zoc_used 
          && !(node_xy->is_my_zoc || node_xy1->is_my_zoc
               || node_xy1->is_allied_tile)) {
        continue;
      }
      
      /* Evaluate the cost of the move */
      cost = (*(the_map->get_cost))(the_map, x1, y1, dir);
      switch (the_map->params->turn_mode) {
      case TM_LUCK:
        if (d_node_xy1->is_dangerous 
            && cost >= get_moves_left(the_map, loc_cost)) {
          /* We would have to end the turn on a dangerous tile! */
          continue;
        }
        break;
      default:
        /* Default is TM_SAFE.  
         * It should be specified explicitly though! */
        if (cost > get_moves_left(the_map, loc_cost) 
            || (d_node_xy1->is_dangerous
                && cost == get_moves_left(the_map, loc_cost))) {
          /* This move is impossible (at least without waiting) */
          continue;
        }
        break;
      }
      
      /* Total cost at xy1 */
      cost += loc_cost;
      /* Is the cost reasonable? */
      if (cost >= PF_MAXCOST) {
        continue;
      } 
      
      /* Evaluate the extra cost of the destination,
       * if it's relevant */
      if (the_map->params->get_ECOT) {
        /* NB: extra_tile is always 0 for dangerous tiles! */
        extra = node_xy1->extra_tile;
        extra += node_xy->extra_cost;
      }
      
      /* Update costs and add to queue, if this is a better route to xy1 */
      if (!d_node_xy1->is_dangerous) {
        int cost_of_path = COP(cost, extra);
        
        if (the_map->status[xy1] == NS_UNINIT
            || (cost_of_path 
                < COP(node_xy1->cost, node_xy1->extra_cost))) {
          node_xy1->extra_cost = extra;
          node_xy1->cost = cost;
          node_xy1->dir_to_here = dir;
          if (d_node_xy->is_dangerous && !d_node_xy1->is_dangerous) {
            /* Transition from red to blue, need to record the path back */
            d_node_xy1->danger_segment
              = create_danger_segment(the_map, dir, &(d_node_xy1->waited));
          } else {
            /* Clear the path back */
            if (d_node_xy1->danger_segment) {
              free(d_node_xy1->danger_segment);
              d_node_xy1->danger_segment = NULL;
            }
            d_node_xy1->waited = (the_map->status[the_map->xy] == NS_WAITING);
          }
          the_map->status[xy1] = NS_NEW;
          pqinsert(the_map->queue, xy1, -cost_of_path);
        }
      } else {
        /* The procedure is slightly different for dangerous nodes */
        /* We will update costs if:
         * 1: we are here for the first time
         * 2: we can possibly go further across dangerous area or
         * 3: we can have lower extra and will not 
         *    overwrite anything useful */
        if (the_map->status[xy1] == NS_UNINIT
            || (get_moves_left(the_map, cost) 
                > get_moves_left(the_map, node_xy1->cost))
            || (COP(cost, extra) 
                < COP(node_xy1->cost, node_xy1->extra_cost)
                && the_map->status[xy1] == NS_PROCESSED)) {
          node_xy1->extra_cost = extra;
          node_xy1->cost = cost;
          node_xy1->dir_to_here = dir;
          the_map->status[xy1] = NS_NEW;
          d_node_xy1->waited = (the_map->status[the_map->xy] == NS_WAITING);
          /* Extra costs of all nodes in danger_queue are equal! */
          pqinsert(the_map->danger_queue, xy1, -cost);
        }
      }
    } adjc_dir_iterate_end;
  }
  
  if (!d_node_xy->is_dangerous && the_map->status[the_map->xy] != NS_WAITING
      && (get_moves_left(the_map, node_xy->cost) 
          < the_map->params->move_rate)) {
    /* Consider waiting at this node. 
     * To do it, put it back into queue. */
    the_map->status[the_map->xy] = NS_WAITING;
    pqinsert(the_map->queue, the_map->xy, 
             -COP(get_moves_left(the_map, node_xy->cost) 
                  + node_xy->cost, node_xy->extra_cost));
  } else {
    the_map->status[the_map->xy] = NS_PROCESSED;
  }  

  /* Get the next nearest node */

  /* First try to get it from danger_queue */
  xy = pqremove(the_map->danger_queue);
  if (xy == PQDATUM_NULL) {
    /* No dangerous nodes to process, go for a safe one */
    do {
      xy = pqremove(the_map->queue);
      if (xy == PQDATUM_NULL) {
        return FALSE;
      }
    } while (the_map->status[the_map->xy] == NS_PROCESSED);
  }
  
  assert(the_map->status[the_map->xy] != NS_UNINIT);

  the_map->xy = xy;
  the_map->x = get_x_from_xy(xy);
  the_map->y = get_y_from_xy(xy);

  if (the_map->status[the_map->xy] == NS_WAITING) {
    /* We've already returned this node once, skip it */
    freelog(LOG_DEBUG, "Considering waiting at (%d, %d)",
            the_map->x, the_map->y);
    return pf_danger_iterate_map(the_map);
  } else if (the_map->d_lattice[xy].is_dangerous) {
    /* We don't return dangerous tiles */
    freelog(LOG_DEBUG, "Reached dangerous tile (%d, %d)",
            the_map->x, the_map->y);
    return pf_danger_iterate_map(the_map);
  } else {
    /* Just return it */
    return TRUE;
  }

#undef COP
}

/*******************************************************************
 * Read off the path to the current node, but with danger
 *******************************************************************/
bool pf_danger_get_current_path(const pf_map_t the_map, 
                                struct pf_path *path)
{
  struct pf_position path_rev[MAX_PATH_LENGTH + 1];
  int i;
  int x = the_map->x, y = the_map->y;
  enum direction8 dir_next = -1;
  struct segment_node *danger_seg = NULL;  /* For danger segments */
  int segment_index;                       /* For danger segments */
  bool waited = FALSE;

  for (i = 0; i <= MAX_PATH_LENGTH - 1; i++) {
    pf_node_t node = &the_map->lattice[get_xy_from_x_and_y(x, y)];
    struct danger_node *d_node 
      = &the_map->d_lattice[get_xy_from_x_and_y(x, y)];
    if (waited && !d_node->is_dangerous) {
      /* Waited at _this_ tile, need to record it twice in the path */
      path_rev[i].x = x;
      path_rev[i].y = y;
      path_rev[i].TEC = node->extra_cost;
      path_rev[i].turn = get_turn(the_map, node->cost) + 1;
      path_rev[i].moves_left = the_map->params->move_rate;
      path_rev[i].cost = path_rev[i].turn * the_map->params->move_rate;
      /* Staying here! */
      path_rev[i].direction_for_next_position = -1;
      i++;
    }

    if (!d_node->is_dangerous) {
      waited = d_node->waited;
    }

    path_rev[i].x = x;
    path_rev[i].y = y;
    if (!d_node->is_dangerous) {
      path_rev[i].cost = node->cost;
      path_rev[i].TEC = node->extra_cost;
    } else if (danger_seg != NULL) {
      path_rev[i].cost = danger_seg[segment_index].cost;
      path_rev[i].TEC = danger_seg[segment_index].extra_cost;
    } else {
      freelog(LOG_ERROR, "PF: No path available on dangerous tile");
    }
    if (the_map->params->turn_mode == TM_SAFE ||
        the_map->params->turn_mode == TM_LUCK) {
      path_rev[i].turn = get_turn(the_map, path_rev[i].cost);
      path_rev[i].moves_left = get_moves_left(the_map, 
                                              path_rev[i].cost);
    } else {
      path_rev[i].turn = -1;
      path_rev[i].moves_left = -1;
    }
    path_rev[i].direction_for_next_position = dir_next;

    if (d_node->danger_segment != NULL) {
      /* We are about to enter danger segment */
      danger_seg = d_node->danger_segment;
      segment_index = 0;
    } else if (!d_node->is_dangerous) {
      /* We left danger segment */
      danger_seg = NULL;
    }

    /* Calculate the next direction */
    if (danger_seg != NULL) {
      /* We are in danger segment */
      dir_next = danger_seg[segment_index].dir;
      segment_index++;
    } else {
      /* We are in normal segment (dir_to_here is valid) */
      dir_next = node->dir_to_here;
    }

    if (same_pos(x, y, 
                 the_map->params->start_x, the_map->params->start_y)) {
      /* Ah-ha, reached the starting point! */
      int j;

      path->positions_used = i + 1;
      path->is_valid = TRUE;
      
      /* Actually filling the path */
      for(j = 0; i>=0; i--, j++) {
        memcpy(&path->positions[j], &path_rev[i], 
               sizeof(struct pf_position));
      }
      
      return TRUE;
    }

    /* Need to step further back */
    {
      int dummy_x, dummy_y;
      
      /* Ugly MAPSTEP... */
      if (!MAPSTEP(dummy_x, dummy_y, x, y, DIR_REVERSE(dir_next))) {
        freelog(LOG_ERROR, "pf_get_current_path: wrong directions recorded!");
        path->is_valid = FALSE;
        return FALSE;
      } else {
        x = dummy_x;
        y = dummy_y;
      }
    }    
  }

  freelog(LOG_NORMAL, "pf_get_current_path: path too long");
  path->is_valid = FALSE;
  return FALSE;
}


/**************************************************************
 * Crap.
 **************************************************************/
void pf_init(void) 
{

  freelog(LOG_NORMAL, "GB: pf_node size: %d", sizeof(struct pf_node));

} 
